package pkg

import (
	"context"
	"errors"
	"net/http"
	"net/url"
	"strconv"
	"strings"
)

const (
	ContextCompanyKey = "numerus-company"
)

type Company struct {
	Id   int
	Slug string
}

func CompanyHandler(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		slug := r.URL.Path
		if idx := strings.IndexByte(slug, '/'); idx >= 0 {
			slug = slug[:idx]
		}

		conn := getConn(r)
		company := &Company{
			Slug: slug,
		}
		err := conn.QueryRow(r.Context(), "select company_id from company where slug = $1", slug).Scan(&company.Id)
		if err != nil {
			http.NotFound(w, r)
			return
		}
		ctx := context.WithValue(r.Context(), ContextCompanyKey, company)
		r = r.WithContext(ctx)

		// Same as StripPrefix
		p := strings.TrimPrefix(r.URL.Path, slug)
		rp := strings.TrimPrefix(r.URL.RawPath, slug)
		if len(p) < len(r.URL.Path) && (r.URL.RawPath == "" || len(rp) < len(r.URL.RawPath)) {
			r2 := new(http.Request)
			*r2 = *r
			r2.URL = new(url.URL)
			*r2.URL = *r.URL
			if p == "" {
				r2.URL.Path = "/"
			} else {
				r2.URL.Path = p
			}
			r2.URL.RawPath = rp
			next.ServeHTTP(w, r2)
		} else {
			http.NotFound(w, r)
		}
	})
}

func getCompany(r *http.Request) *Company {
	company := r.Context().Value(ContextCompanyKey)
	if company == nil {
		return nil
	}
	return company.(*Company)
}

type CurrencyOption struct {
	Code   string
	Symbol string
}

type CountryOption struct {
	Code string
	Name string
}

type Tax struct {
	Id   int
	Name string
	Rate int
}

type TaxDetailsPage struct {
	Title        string
	BusinessName string
	VATIN        string
	TradeName    string
	Phone        string
	Email        string
	Web          string
	Address      string
	City         string
	Province     string
	PostalCode   string
	CountryCode  string
	Countries    []CountryOption
	CurrencyCode string
	Currencies   []CurrencyOption
	Taxes        []Tax
}

func CompanyTaxDetailsHandler() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		locale := getLocale(r)
		page := &TaxDetailsPage{
			Title: pgettext("title", "Tax Details", locale),
		}
		company := mustGetCompany(r)
		conn := getConn(r)
		if r.Method == "POST" {
			r.ParseForm()
			page.BusinessName = r.FormValue("business_name")
			page.CountryCode = r.FormValue("country")
			page.VATIN = page.CountryCode + r.FormValue("vatin")
			page.TradeName = r.FormValue("trade_name")
			page.Phone = r.FormValue("phone")
			page.Email = r.FormValue("email")
			page.Web = r.FormValue("web")
			page.Address = r.FormValue("address")
			page.City = r.FormValue("city")
			page.Province = r.FormValue("province")
			page.PostalCode = r.FormValue("postal_code")
			page.CurrencyCode = r.FormValue("currency")
			conn.MustExec(r.Context(), "update company set business_name = $1, vatin = $2, trade_name = $3, phone = parse_packed_phone_number($4, $11), email = $5, web = $6, address = $7, city = $8, province = $9, postal_code = $10, country_code = $11, currency_code = $12 where company_id = $13", page.BusinessName, page.VATIN, page.TradeName, page.Phone, page.Email, page.Web, page.Address, page.City, page.Province, page.PostalCode, page.CountryCode, page.CurrencyCode, company.Id)
			http.Redirect(w, r, "/company/"+company.Slug+"/tax-details", http.StatusSeeOther)
		} else {
			err := conn.QueryRow(r.Context(), "select business_name, substr(vatin::text, 3), trade_name, phone, email, web, address, city, province, postal_code, country_code, currency_code from company where company_id = $1", company.Id).Scan(&page.BusinessName, &page.VATIN, &page.TradeName, &page.Phone, &page.Email, &page.Web, &page.Address, &page.City, &page.Province, &page.PostalCode, &page.CountryCode, &page.CurrencyCode)
			if err != nil {
				panic(err)
			}
		}
		page.Countries = mustGetCountryOptions(r.Context(), conn, locale)
		page.Currencies = mustGetCurrencyOptions(r.Context(), conn)
		page.Taxes = mustGetTaxes(r.Context(), conn, company)
		mustRenderAppTemplate(w, r, "tax-details.html", page)
	})
}

func mustGetCompany(r *http.Request) *Company {
	company := getCompany(r)
	if company == nil {
		panic(errors.New("company: required but not found"))
	}
	return company
}

func mustGetCountryOptions(ctx context.Context, conn *Conn, locale *Locale) []CountryOption {
	rows, err := conn.Query(ctx, "select country.country_code, coalesce(i18n.name, country.name) as l10n_name from country left join country_i18n as i18n on country.country_code = i18n.country_code and i18n.lang_tag = $1 order by l10n_name", locale.Language)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var countries []CountryOption
	for rows.Next() {
		var country CountryOption
		err = rows.Scan(&country.Code, &country.Name)
		if err != nil {
			panic(err)
		}
		countries = append(countries, country)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return countries
}

func mustGetCurrencyOptions(ctx context.Context, conn *Conn) []CurrencyOption {
	rows, err := conn.Query(ctx, "select currency_code, currency_symbol from currency order by currency_code")
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var currencies []CurrencyOption
	for rows.Next() {
		var currency CurrencyOption
		err = rows.Scan(&currency.Code, &currency.Symbol)
		if err != nil {
			panic(err)
		}
		currencies = append(currencies, currency)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return currencies
}

func mustGetTaxes(ctx context.Context, conn *Conn, company *Company) []Tax {
	rows, err := conn.Query(ctx, "select tax_id, name, (rate * 100)::integer from tax where company_id = $1 order by rate, name", company.Id)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var taxes []Tax
	for rows.Next() {
		var tax Tax
		err = rows.Scan(&tax.Id, &tax.Name, &tax.Rate)
		if err != nil {
			panic(err)
		}
		taxes = append(taxes, tax)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return taxes
}

func CompanyTaxHandler() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		param := r.URL.Path
		if idx := strings.LastIndexByte(param, '/'); idx >= 0 {
			param = param[idx+1:]
		}
		conn := getConn(r)
		company := mustGetCompany(r)
		if taxId, err := strconv.Atoi(param); err == nil {
			conn.MustExec(r.Context(), "delete from tax where tax_id = $1", taxId)
		} else {
			r.ParseForm()
			name := r.FormValue("name")
			rate, _ := strconv.Atoi(r.FormValue("rate"))
			conn.MustExec(r.Context(), "insert into tax (company_id, name, rate) values ($1, $2, $3 / 100::decimal)", company.Id, name, rate)
		}
		http.Redirect(w, r, "/company/"+company.Slug+"/tax-details", http.StatusSeeOther)
	})
}
